from .common import *


def get_action_tags(istack, stack_tags):
    # unchanged tags
    tags_default = {}

    # changed tags - same value as corresponding stack param
    tags_changed = {}
    final_tags = []

    for tag in stack_tags:
        key = tag['Key']
        current_value = tag['Value']

        # check if key exist as cfg param/attr too
        try:
            cfg_value = getattr(istack.cfg, key)
            in_cfg = True if cfg_value is not None else None
        except Exception:
            in_cfg = None

        # Skip LastUpdate Tag
        if key == 'LastUpdate':
            continue

        # current value differ from cmd arg
        if in_cfg and current_value != cfg_value:
            value = cfg_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_changed[key] = '%s => %s' % (current_value, value)

        # keep current tag value
        else:
            value = current_value

            # tags value cannot be empty
            if len(value) == 0:
                value = "empty"

            tags_default[key] = value

        final_tags.append({
            'Key': key,
            'Value': value
        })

    # Add LastUpdate Tag with current time
    final_tags.append({
        'Key': 'LastUpdate',
        'Value': str(datetime.now())
    })

    if len(tags_default) > 0:
        istack.mylog(
            'DEFAULT - STACK TAGS\n%s\n' % pformat(
                tags_default, width=1000000))
    if len(tags_changed) > 0:
        istack.mylog(
            'CHANGED - STACK TAGS\n%s\n' % pformat(
                tags_changed, width=1000000))

    return final_tags



# parser default cfg
stack = role = type = topics = stack_args = []
parallel = region = jobs = pause = version = template = None
nowait = compact = dryrun = answer_yes = no_stacks = all_stacks = None
debug = False
max_retry_ecs_service_running_count = 0
timedelta = 300
dashboard = 'OnChange'
statistic = 'Average'
statisticresponse = 'p95'
silent = True
vertical = False
#

OUT_WIDTH = 1000000

SLACK_CHANNEL = '_cf_deploy'

MAX_SINGLE_STACKS = 5

ACTION_WAITER_SLEEP_TIME = 3

STACK_BASE_DATA = [
    'StackName',
    'Description',
    'StackStatus',
    'CreationTime',
    'LastUpdatedTime',
]

RESOURCES_MAP = {
    'AutoScalingGroup': 'AutoScalingGroupName',
    'AutoScalingGroupSpot': 'AutoScalingGroupSpotName',
    'TargetGroup': 'TargetGroup',
    'TargetGroupExternal': 'TargetGroupExternal',
    'TargetGroupInternal': 'TargetGroupInternal',
    'Service': 'ServiceName',
    'ServiceExternal': 'ServiceName',
    'ServiceInternal': 'ServiceName',
    'LoadBalancerClassicExternal': 'LoadBalancerNameExternal',
    'LoadBalancerClassicInternal': 'LoadBalancerNameInternal',
    'LoadBalancerApplicationExternal': 'LoadBalancerExternal',
    'LoadBalancerApplicationInternal': 'LoadBalancerInternal',
    'Cluster': 'ClusterName',
    'ScalableTarget': 'ClusterName',
    'ListenerHttpsExternalRules1': 'LoadBalancerExternal',
    'ListenerHttpsExternalRules2': 'LoadBalancerExternal',
    'ListenerHttpExternalRules1': 'LoadBalancerExternal',
    'ListenerHttpInternalRules1': 'LoadBalancerInternal',
    'AlarmCPUHigh': None,
    'AlarmCPULow': None,
}
SCALING_POLICY_TRACKINGS_NAMES = {
    'ScalingPolicyTrackings1': None,
    'ScalingPolicyTrackingsASCpu': 'ScalingPolicyTrackings1',
    'ScalingPolicyTrackingsASCustom': 'ScalingPolicyTrackings1',
    'ScalingPolicyTrackingsAPPCpu': 'ScalingPolicyTrackings1',
    'ScalingPolicyTrackingsAPPCustom': 'ScalingPolicyTrackings1',
    'AutoScalingScalingPolicyCpu': 'ScalingPolicyTrackings1',
    'AutoScalingScalingPolicyCustom': 'ScalingPolicyTrackings1',
    'ApplicationAutoScalingScalingPolicyCpu': 'ScalingPolicyTrackings1',
    'ApplicationAutoScalingScalingPolicyCustom': 'ScalingPolicyTrackings1',

    # Disabled until they are properly managed in dashboard.py [get_policy].
    # Ideal should be to create two horizontal annotations
    # one for the lower bound and one for the upper one
    # but as AutoScalingScalingPolicy have been replaced with Tracking ones
    # i do not know if it make sense to put much effort in this.

    # 'AutoScalingScalingPolicyDown': 'ScalingPolicyTrackings1',
    # 'AutoScalingScalingPolicyUp': 'ScalingPolicyTrackings1',
    # 'ApplicationAutoScalingScalingPolicyDown': 'ScalingPolicyTrackings1',
    # 'ApplicationAutoScalingScalingPolicyUp': 'ScalingPolicyTrackings1',
}
RESOURCES_MAP.update(SCALING_POLICY_TRACKINGS_NAMES)

RESOURCES_MAP_R53 = {
    'RecordSetExternal': None,
    'RecordSetInternal': None,
    'RecordSetExternalRO': None,
    'RecordSetInternalRO': None,
    'RecordSetCloudFront': None,
}

STACK_COMPLETE_STATUS = [
    'UPDATE_COMPLETE',
    'CREATE_COMPLETE',
    'ROLLBACK_COMPLETE',
    'UPDATE_ROLLBACK_COMPLETE',
    'UPDATE_ROLLBACK_FAILED',
    'DELETE_COMPLETE',
    'DELETE_FAILED',
]

CHANGESET_COMPLETE_STATUS = [
    'CREATE_COMPLETE',
    'UPDATE_ROLLBACK_FAILED',
    'FAILED',
]

SHOW_TABLE_FIELDS = [
    'EnvStackVersion',
    'EnvRole',
    'StackName',
    'StackType',
    'UpdateMode',
    'EnvApp1Version',
    'StackStatus',
    'LastUpdatedTime',
]

SSM_BASE_PATH = '/ibox'



import argparse
from . import cfg
from .commands import (create, update, delete, cancel_update, continue_update,
                       info, parameters, resolve, show_table, log, dash,
                       ssm_setup, ssm_put, ssm_show, r53, replicate)


def get_create_parser(subparser, parents=[]):
    parser = subparser.add_parser('create',
                                  parents=parents,
                                  help='Create Stack')
    parser.add_argument('--Env',
                        help='Environment to use',
                        type=str, required=True)
    parser.add_argument('--EnvRole',
                        help='Stack Role',
                        type=str, required=True)
    parser.add_argument('--EnvApp1Version',
                        help='App Version',
                        type=str, default='')

    return parser


def get_update_parser(subparser, parents=[]):
    parser = subparser.add_parser('update',
                                  parents=parents,
                                  help='Update Stack')
    parser.add_argument('-P', '--policy',
                        help='Policy during Stack Update',
                        type=str, choices=[
                            '*', 'Modify', 'Delete', 'Replace',
                            'Modify,Delete', 'Modify,Replace',
                            'Delete,Replace'])
    parser.add_argument('-n', '--nochangeset',
                        help='No ChangeSet',
                        required=False, action='store_true')
    parser.add_argument('--dryrun',
                        help='Show changeset and exit',
                        action='store_true')
    parser.add_argument('-T', '--showtags',
                        help='Show tags changes in changeset',
                        action='store_true')
    parser.add_argument('-D', '--dashboard',
                        help='Update CloudWatch DashBoard',
                        choices=[
                            'Always', 'OnChange', 'Generic', 'None'],
                        default=cfg.dashboard)
    parser.add_argument('--nodetails',
                        help='Do not show extra details in changeset',
                        action='store_true')

    return parser


def set_dash_parser(subparser, parents=[]):
    parser = subparser.add_parser('dash',
                                  parents=parents,
                                  help='Create DashBoard for stacks')
    parser.set_defaults(func=dash)

    parser.add_argument(
        '--statistic',
        help='Statistic to use for metrics',
        choices=['Average', 'Maximum', 'Minimum'],
        default='Average'
    )
    parser.add_argument(
        '--statisticresponse',
        help='Statistic to use for response time metrics',
        choices=[
            'Average', 'p99', 'p95', 'p90',
            'p50', 'p10', 'Maximum', 'Minimum'],
        default='p95',
    )
    parser.add_argument('--debug', help='Show json Dash', action='store_true')
    parser.add_argument('--silent', help='Silent mode', action='store_true')
    parser.add_argument(
        '--vertical',
        help='Add vertical annotation at creation time, '
             'and optionally specify fill mode',
        nargs='?',
        choices=['before', 'after'],
        const=True,
        default=False,
    )


def get_show_parser(subparser, parents=[]):
    parser = subparser.add_parser('show',
                                  parents=parents,
                                  help='Show Stacks table')

    parser.add_argument('-F', '--fields', nargs='+',
                        type=str, default=cfg.SHOW_TABLE_FIELDS)
    parser.add_argument('-O', '--output',
                        type=str, default='text',
                        choices=['text', 'html', 'bare'])

    return parser


def set_ssm_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        'ssm',
        parents=[],
        help='SSM Parameters override for Stack Replicator')

    ssm_parser = parser.add_subparsers(title='SSM Command',
                                       required=True, dest='command_ssm')
    # setup
    setup_parser = ssm_parser.add_parser(
        'setup', help='Setup Regions',
        parents=parents)
    setup_parser.set_defaults(func=ssm_setup, no_stacks=True)

    setup_parser.add_argument('-R', '--regions',
                              help='Regions', type=str,
                              required=True, default=[], nargs='+')
    # put
    put_parser = ssm_parser.add_parser(
        'put', help='Put Parameters',
        parents=parents)
    put_parser.set_defaults(func=ssm_put)

    put_parser.add_argument('-R', '--regions',
                            help='Regions', type=str,
                            default=[], nargs='+')
    # show
    show_parser = ssm_parser.add_parser(
        'show', help='Show Regions Distribution',
        parents=parents)
    show_parser.set_defaults(func=ssm_show, all_stacks=True)


def set_replicate_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        'replicate',
        parents=[],
        help='Replicate in Regions configured by SSM')
    parser.set_defaults(func=replicate)

    replicate_parser = parser.add_subparsers(title='Replicate Command',
                                             required=True,
                                             dest='command_replicate')

    regions_parser = argparse.ArgumentParser(add_help=False)
    regions_parser.add_argument('-R', '--regions',
                                help='Regions', type=str,
                                default=[], nargs='+')
    regions_parser.add_argument('--no_replicate_current',
                                help='No replication in current Region',
                                action='store_true')

    # replicate create
    parser_create = get_create_parser(
        replicate_parser, parents + [
            get_template_parser(),
            get_stack_single_parser(),
            get_create_update_parser(),
            regions_parser])

    # replicate update
    parser_update = get_update_parser(
        replicate_parser, parents + [
            get_template_parser(required=False),
            get_stack_selection_parser(),
            get_create_update_parser(),
            regions_parser])

    # replicate delete
    parser_delete = replicate_parser.add_parser(
        'delete',
        parents=parents + [
            get_stack_single_parser(),
            regions_parser],
        help='Delete Stack (WARNING)')

    # replicate show
    parser_show = get_show_parser(
        replicate_parser, [
            get_stack_selection_parser(),
            regions_parser])


def set_r53_parser(subparser, parents=[]):
    parser = subparser.add_parser(
        'r53', parents=parents,
        help='Create RecordSet Aliases looking at stack R53 resources')
    parser.set_defaults(func=r53)

    parser.add_argument('--dryrun', help='Show changes and exit',
                        action="store_true")
    parser.add_argument('--noorigin', help='Do not create Origin Record',
                        action="store_true")
    parser.add_argument(
        '--suffix',
        help='Suffix to add to RecordSet Name, prepended with "-"',
        default='', type=str)


def get_template_parser(required=True):
    parser = argparse.ArgumentParser(add_help=False)

    group = parser.add_mutually_exclusive_group(required=required)
    group.add_argument('--template',
                       help='Template Location',
                       type=str)
    group.add_argument('-v', '--version',
                       help='Stack Env Version',
                       type=str)

    return parser


def get_stack_selection_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        '-s', '--stack', nargs='+',
        help='Stack Names space separated',
        type=str, default=[])
    parser.add_argument(
        '-r', '--role', nargs='+',
        help='Stack Roles space separated',
        type=str, default=[])
    parser.add_argument(
        '-t', '--type', nargs='+',
        help='Stack Types space separated - use ALL for any type',
        type=str, default=[])

    return parser


def get_stack_single_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        '-s', '--stack', nargs=1,
        help='Stack Names space separated',
        required=True, type=str, default=[])

    return parser


def get_create_update_parser():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument('--topics', nargs='+',
                        help='SNS Topics Arn for notification',
                        type=str, default=[])
    parser.add_argument('-M', '--max_retry_ecs_service_running_count',
                        help='Max retry numbers when updating ECS '
                             'service and runningCount is stuck to zero',
                        type=int, default=0)

    return parser


# parse main argumets
def get_parser():
    parser = argparse.ArgumentParser(
        description='Stacks Operations',
        epilog='Note: options for Stack Params must be put at the end!'
    )

    # common parser
    parser.add_argument(
        '--region',
        help='Region', type=str)
    parser.add_argument(
        '--profile',
        help='AWS Profile', type=str)
    parser.add_argument(
        '--compact',
        help='Display Stacks-Output in compact form',
        action='store_true')
    parser.add_argument(
        '-j', '--jobs',
        help='Max Concurrent jobs - default to number of stacks', type=int)
    parser.add_argument(
        '--pause',
        help='Pause for seconds between jobs - '
             '0 for interactive - valid only for jobs=1',
        type=int)

    # action parser
    action_parser = argparse.ArgumentParser(add_help=False)

    action_parser.add_argument('-y', '--answer_yes',
                               help='Answer YES (No Confirm)',
                               required=False, action='store_true')
    action_parser.add_argument('-w', '--nowait',
                               help='Do not Wait for action to end',
                               required=False, action='store_true')
    action_parser.add_argument('-c', '--slack_channel',
                               help=f'Slack Channel [{cfg.SLACK_CHANNEL}]',
                               nargs='?', const=cfg.SLACK_CHANNEL,
                               default=False)

    # template parser
    template_parser_create = get_template_parser()
    template_parser_update = get_template_parser(required=False)

    # stack selection parser
    stack_selection_parser = get_stack_selection_parser()

    # stack single parser
    stack_single_parser = get_stack_single_parser()

    # create update parser
    create_update_parser = get_create_update_parser()

    # command subparser
    command_subparser = parser.add_subparsers(
        help='Desired Command',
        required=True,
        dest='command')

    # create parser
    parser_create = get_create_parser(
        command_subparser, [
            action_parser,
            template_parser_create,
            stack_single_parser,
            create_update_parser,
        ])
    parser_create.set_defaults(func=create)

    # update parser
    parser_update = get_update_parser(
        command_subparser, [
            action_parser,
            template_parser_update,
            stack_selection_parser,
            create_update_parser,
        ])
    parser_update.set_defaults(func=update)

    # delete parser
    parser_delete = command_subparser.add_parser(
        'delete',
        parents=[
            action_parser,
            stack_single_parser],
        help='Delete Stack (WARNING)')
    parser_delete.set_defaults(func=delete)

    # cancel_update parser
    parser_cancel = command_subparser.add_parser(
        'cancel',
        parents=[
            action_parser,
            stack_selection_parser],
        help='Cancel Update Stack')
    parser_cancel.set_defaults(func=cancel_update)

    # continue_update parser
    parser_continue = command_subparser.add_parser(
        'continue',
        parents=[
            action_parser,
            stack_selection_parser],
        help='Continue Update RollBack')
    parser_continue.set_defaults(func=continue_update)
    parser_continue.add_argument(
        '--resources_to_skip',
        help='Resource to Skip',
        default=[], nargs='+')

    # info parser
    parser_info = command_subparser.add_parser(
        'info', parents=[stack_selection_parser],
        help='Show Stack Info')
    parser_info.set_defaults(func=info)

    # parameters parser
    parser_parameters = command_subparser.add_parser(
        'parameters', parents=[
            template_parser_update,
            stack_selection_parser],
        help='Show Available Stack Parameters')
    parser_parameters.set_defaults(func=parameters)

    # resolve parser
    parser_resolve = command_subparser.add_parser(
        'resolve', parents=[
            template_parser_update,
            stack_selection_parser],
        help='Resolve Stack template - output in yaml short format')
    parser_resolve.set_defaults(func=resolve)

    # log parser
    parser_log = command_subparser.add_parser(
        'log',
        parents=[
            stack_single_parser],
        help='Show Stack Log')
    parser_log.set_defaults(func=log)
    parser_log.add_argument(
        '-d', '--timedelta',
        help='How many seconds go back in time from stack last event - '
             'use 0 for realtime - if < 30 assume days',
        default=cfg.timedelta)

    # dashboard parser
    set_dash_parser(
        command_subparser, [
            stack_selection_parser,
        ])

    # show parser
    parser_show = get_show_parser(
        command_subparser, [
            stack_selection_parser,
        ])
    parser_show.set_defaults(func=show_table, all_stacks=True)

    # ssm parser
    set_ssm_parser(
        command_subparser, [
            stack_selection_parser,
        ])

    # replicate parser
    set_replicate_parser(
        command_subparser, [
            action_parser,
        ])

    # r53 parser
    set_r53_parser(
        command_subparser, [
            stack_selection_parser,
        ])

    return parser


def set_cfg(argv):
    parser = get_parser()
    args = parser.parse_known_args(argv)

    for n, v in vars(args[0]).items():
        setattr(cfg, n, v)

    # trick for showing ALL Stacks
    # if nor stack nor role nor type are specified.
    if (cfg.all_stacks
            and not (cfg.stack or cfg.role or cfg.type)):
        cfg.type = ['ALL']

    if (not cfg.no_stacks
            and not (cfg.stack or cfg.role or cfg.type)):
        parser.print_help()
        exit(0)

    cfg.stack_args = args[1]



import logging
from . import cfg
from .common import *

try:
    import slack
    HAVE_SLACK = True
except ModuleNotFoundError:
    HAVE_SLACK = False

logging.basicConfig()
logging.getLogger('botocore').setLevel('CRITICAL')
logger = logging.getLogger('stacksops')
logger.setLevel(logging.INFO)


def get_msg_client():
    try:
        cfg.slack_channel
    except Exception:
        cfg.slack_channel = None

    if (cfg.slack_channel
            and HAVE_SLACK
            and 'IBOX_SLACK_TOKEN' in os.environ
            and 'IBOX_SLACK_USER' in os.environ):
        slack_web = slack.WebClient(token=os.environ['IBOX_SLACK_TOKEN'])
        return slack_web

    return None



from prettytable import PrettyTable, ALL as ptALL
from .tools import show_confirm
from .common import *


# create changeset
def _do_changeset(istack, us_args):
    if not istack.cfg.showtags:
        # keep existing stack.tags so that they are excluded from changeset
        # (not the new prepared ones)
        us_args['Tags'] = istack.stack.tags

    us_args['StackName'] = istack.name
    us_args['ChangeSetName'] = (
        'CS-%s' % time.strftime('%Y-%m-%dT%H-%M-%S')
    )
    us_args.pop('StackPolicyDuringUpdateBody', None)

    response = istack.client.create_change_set(**us_args)

    return response['Id']


# wait until changeset is created
def _changeset_waiter(istack, changeset_id):
    while True:
        time.sleep(3)
        changeset = istack.client.describe_change_set(
            ChangeSetName=changeset_id, StackName=istack.name)
        if changeset['Status'] in istack.cfg.CHANGESET_COMPLETE_STATUS:
            return changeset


# parse changeset changes
def _parse_changeset(changeset):
    changes = []
    for change in changeset['Changes']:
        change_dict = {}
        ResChange = change['ResourceChange']
        change_dict['LogicalResourceId'] = ResChange['LogicalResourceId']
        change_dict['Action'] = ResChange['Action']
        change_dict['ResourceType'] = ResChange['ResourceType']
        if ResChange['Action'] == 'Modify':
            change_dict['Replacement'] = ResChange['Replacement']
            scope = ResChange['Scope']
            change_dict['Scope'] = list_to_string_list(scope)
            target = []
            causingentity = []
            for i in ResChange['Details']:
                if 'Name' in i['Target'] and i['Target']['Name'] not in target:
                    target.append(i['Target']['Name'])
                if ('CausingEntity' in i and
                        'Name' in i['Target'] and
                        i['CausingEntity'] not in causingentity):
                    causingentity.append(i['CausingEntity'])
            change_dict['Target'] = list_to_string_list(target)
            change_dict['CausingEntity'] = list_to_string_list(causingentity)

        changes.append(change_dict)
    return changes


def _show_changeset_changes(istack, changes):
    fields = ['LogicalResourceId', 'ResourceType', 'Action']
    fileds_ex = ['Replacement', 'Scope', 'Target', 'CausingEntity']
    fields.extend(fileds_ex)
    table = PrettyTable()
    if istack.cfg.nodetails:
        fields.remove('Target')
        fields.remove('CausingEntity')
    table.field_names = fields
    table.padding_width = 1
    table.align['LogicalResourceId'] = "l"
    table.align['ResourceType'] = "l"
    for row in changes:
        table.add_row([
            'None' if i in fileds_ex and row['Action'] != 'Modify'
            else row[i] for i in fields])

    istack.mylog('ChangeSet:')
    print(table.get_string(fields=fields))


def _delete_changeset(istack, changeset_id):
    response = istack.client.delete_change_set(
        ChangeSetName=changeset_id,
        StackName=istack.name
    )

    return response


def _execute_changeset(istack, changeset_id):
    response = istack.client.execute_change_set(
        ChangeSetName=changeset_id,
        StackName=istack.name
    )

    return response


# change list in a string new line separated element
def list_to_string_list(mylist):
    joined_string = '\n'.join(mylist)
    mystring = joined_string if len(mylist) < 2 else f'({joined_string})'

    return mystring


def process(istack, us_args):
    # -create changeset
    changeset_id = _do_changeset(istack, us_args.copy())
    print('\n')
    istack.mylog('ChangeSetId: %s' % changeset_id)
    print('\n')
    time.sleep(1)
    istack.mylog('Waiting ChangeSet Creation..')

    # -wait changeset creation and return it
    changeset = _changeset_waiter(istack, changeset_id)
    # pprint(changeset)

    # -parse changeset changes
    changeset_changes = _parse_changeset(changeset)

    # -show changeset changes
    _show_changeset_changes(istack, changeset_changes)

    # -delete changeset
    _delete_changeset(istack, changeset_id)

    if not istack.cfg.dryrun and show_confirm():
        # _execute_changeset(istack, changeset_id)
        return True
    else:
        return None
        # _delete_changeset(istack, changeset_id)



from collections import OrderedDict
from .log import logger, get_msg_client
from .common import *


def _update_template_param(istack):
    # try to get role from istack.cfg or use current stack parameter value
    try:
        role = istack.cfg.EnvRole
    except Exception:
        role = istack.c_parameters['EnvRole']

    app_repository = istack.exports['BucketAppRepository']
    s3_prefix = 'ibox/%s/templates/%s.' % (istack.cfg.version, role)

    try:
        response = istack.s3.list_objects_v2(
            Bucket=app_repository, Prefix=s3_prefix)
        istack.cfg.template = 'https://%s.s3.amazonaws.com/%s' % (
            app_repository, response['Contents'][0]['Key'])
    except Exception:
        raise IboxError(
            f'Error retrieving stack template with prefix: {s3_prefix}')


def get_template(istack):
    logger.info('Getting Template Body')
    # update template param if using version one
    if istack.cfg.version:
        _update_template_param(istack)

    try:
        # New template
        if istack.cfg.template:
            template = str(istack.cfg.template)
            # From s3
            if template.startswith('https'):
                url = template[template.find('//') + 2:]
                bucket = url[:url.find('.')]
                key = url[url.find('/') + 1:]

                response = istack.s3.get_object(Bucket=bucket, Key=key)
                body = response['Body'].read()
                istack.template_from = 'S3'
            # From file
            else:
                f = open(istack.cfg.template[7:], 'r')
                body = f.read()
                istack.template_from = 'File'
        # Current template
        else:
            response = istack.client.get_template(StackName=istack.name)
            tbody = response['TemplateBody']
            if isinstance(tbody, OrderedDict):
                body = json.dumps(tbody)
            else:
                body = tbody
            istack.template_from = 'Current'

    except Exception as e:
        raise IboxError(f'Error retrieving template: {e}')
    else:
        try:
            template = json.loads(body)
        except Exception:
            try:
                template = yaml.load(body, Loader=yaml.FullLoader)
            except Exception:
                raise IboxError('Error parsing template body')

        for n in ['Parameters', 'Conditions', 'Mappings', 'Resources']:
            setattr(istack, n.lower(), template.get(n))

        return template



from collections import OrderedDict
from .log import logger
from .common import *


def _process_template(istack):

    def _recursive_resolve(name, value):
        if isinstance(value, (dict, OrderedDict)):
            r_root = {}
            for r, v in value.items():
                if r == 'Fn::If':
                    return _resolve_if(name, v)
                elif r == 'Ref':
                    return _resolve_ref(name, v)
                elif r == 'Fn::GetAtt':
                    return '!GetAtt %s.%s' % (v[0], v[1])
                elif r == 'Fn::GetAZs':
                    return '!GetAZs %s' % istack.boto3.region_name
                elif r == 'Fn::ImportValue':
                    return _resolve_import(name, v)
                elif r == 'Fn::Sub':
                    return _resolve_sub(name, v)
                elif r == 'Fn::FindInMap':
                    return _resolve_findinmap(name, v)
                elif r == 'Fn::Join':
                    return _resolve_join(name, v)
                elif r == 'Fn::Select':
                    return _resolve_select(name, v)
                elif r == 'Fn::Split':
                    return _resolve_split(name, v)
                elif r == 'Fn::Or':
                    return _resolve_or(name, v)
                elif r == 'Fn::And':
                    return _resolve_and(name, v)
                elif r == 'Fn::Equals':
                    return _resolve_equals(name, v)
                elif r == 'Fn::Not':
                    return _resolve_not(name, v)
                elif r == 'Condition' and isinstance(v, str):
                    return _resolve_condition(name, v)
                else:
                    r_value = _recursive_resolve(name, v)

                    if not _awsnovalue(r_value):
                        r_root[r] = r_value

            return r_root

        elif isinstance(value, list):
            r_root = []
            for n, l in enumerate(value):
                r_value = _recursive_resolve(n, l)

                if not _awsnovalue(r_value):
                    r_root.append(r_value)

            return r_root

        elif isinstance(value, (int, str)):

            return value

    def _resolve_sub(name, value):
        if isinstance(value, list):
            sub_string = value[0]
            sub_data = value[1]
        else:
            sub_string = value
            sub_data = ''

        while True:
            found = sub_string.find('${')
            if found == -1:
                break
            find_start = found + 2
            find_end = sub_string[find_start:].find('}') + find_start
            key = sub_string[find_start:find_end]
            replace_from = '${' + key + '}'

            if key in sub_data:
                r_value = _recursive_resolve(key, sub_data[key])
                replace_to = r_value
            elif key in istack.r_parameters:
                replace_to = istack.r_parameters[key]
            elif key == 'AWS::Region':
                replace_to = istack.boto3.region_name
            elif key == 'AWS::AccountId':
                replace_to = istack.boto3.client(
                    'sts').get_caller_identity()['Account']
            else:
                replace_to = key

            sub_string = sub_string.replace(replace_from, str(replace_to))

        if sub_string.startswith('https://'):
            _find_s3_files(name, sub_string)

        return sub_string

    def _resolve_if(name, v):
        value = v[1] if istack.r_conditions[v[0]] else v[2]

        return _recursive_resolve(name, value)

    def _resolve_ref(name, v):
        if v == 'AWS::Region':
            value = istack.boto3.region_name
        elif v in istack.r_parameters:
            value = istack.r_parameters[v]
        else:
            # value = {'Ref': v}
            value = f'!Ref {v}'

        return value

    def _resolve_import(name, v):
        value = _recursive_resolve(name, v)

        return istack.exports[value]

    def _resolve_findinmap(name, v):
        mapname = _recursive_resolve(name, v[0])
        key = _recursive_resolve(name, v[1])
        key_value = _recursive_resolve(name, v[2])

        value = istack.mappings[mapname][key][key_value]

        if isinstance(value, list):
            return value
        else:
            return str(value)

        return str(istack.mappings[mapname][key][key_value])

    def _resolve_join(name, v):
        j_list = []
        for n in v[1]:
            j_list.append(str(_recursive_resolve(name, n)))

        return v[0].join(j_list)

    def _resolve_select(name, v):
        index = v[0]
        s_list = _recursive_resolve(name, v[1])

        try:
            value = s_list[index]
        except Exception:
            value = s_list

        return value

    def _resolve_split(name, v):
        delimeter = v[0]
        s_string = _recursive_resolve(name, v[1])

        return s_string.split(delimeter)

    def _resolve_or(name, v):
        o_list = []
        for n in v:
            o_list.append(_recursive_resolve(name, n))

        return any(o_list)

    def _resolve_and(name, v):
        o_list = []
        for n in v:
            o_list.append(_recursive_resolve(name, n))

        return all(o_list)

    def _resolve_equals(name, v):
        first_value = str(_recursive_resolve(name, v[0]))
        second_value = str(_recursive_resolve(name, v[1]))

        return True if first_value == second_value else False

    def _resolve_not(name, v):
        value = True if not _recursive_resolve(name, v[0]) else False

        return value

    def _resolve_condition(name, v):
        istack.r_conditions[v] = _recursive_resolve(name, istack.conditions[v])

        return istack.r_conditions[v]

    def _awsnovalue(value):
        if value == 'AWS::NoValue' or value == '!Ref AWS::NoValue':
            return True

        return False

    def _find_s3_files(name, sub_string):
        if ('AWS::AutoScaling::LaunchConfiguration' in istack.t_resources and
                istack.t_resources[
                    'AWS::AutoScaling::LaunchConfiguration'] == name):
            data = sub_string[8:].partition('/')
            host = data[0]
            if not host.endswith('amazonaws.com'):
                return
            key = data[2]
            s_bucket = host.rsplit('.', 3)
            if s_bucket[1].startswith('s3-'):
                bucket = s_bucket[0]
            else:
                bucket = host[:host.rfind('.s3.')]

            if host == bucket:
                return

            istack.s3_files.add((bucket, key))

    def _check_lambda(r):
        try:
            bucket = istack.r_resources[r]['Properties']['Code']['S3Bucket']
            key = istack.r_resources[r]['Properties']['Code']['S3Key']
        except Exception:
            pass
        else:
            istack.s3_files.add((bucket, key))

    def _process_conditions():
        istack.r_conditions = {}
        for c in sorted(istack.conditions):
            v = istack.conditions[c]
            if c not in istack.r_conditions:
                istack.r_conditions[c] = _recursive_resolve(c, v)

    def _process_resources():
        istack.r_resources = {}
        istack.t_resources = {}
        istack.s3_files = set()

        for r in sorted(istack.resources):
            v = istack.resources[r]
            if not ('Condition' in v
                    and not istack.r_conditions[v['Condition']]):
                try:
                    del v['Condition']
                except Exception:
                    pass
                istack.t_resources[v['Type']] = r
                istack.r_resources[r] = _recursive_resolve(r, v)

                if v['Type'] == 'AWS::Lambda::Function':
                    _check_lambda(r)

    _process_conditions()
    _process_resources()


def _check_s3_files(istack):
    for f in istack.s3_files:
        bucket = f[0]
        key = f[1]
        try:
            istack.s3.head_object(Bucket=bucket, Key=key)
        except botocore.exceptions.ClientError as e:
            logger.error(f'Missing: {bucket}/{key}')
            raise IboxError(e)


def _check_ecr_images(istack):
    name = istack.t_resources['AWS::ECS::TaskDefinition']
    images = []
    ecr = istack.boto3.client('ecr')

    for c in (istack.r_resources[name]
              ['Properties']['ContainerDefinitions']):
        image = c['Image']
        registry_id = image[0:image.find('.')]
        repository_name = image[image.find('/')+1:image.find(':')]
        image_id = image[image.find(':') + 1:]
        # Skip already processed images and images from public docker repo
        if (image not in images and
                '.dkr.ecr.' in image):
            try:
                ecr.describe_images(
                    registryId=registry_id,
                    repositoryName=repository_name,
                    imageIds=[{
                        'imageTag': image_id,
                    }],
                )
                images.append(image)
            except botocore.exceptions.ClientError as e:
                logger.error(f'Missing: {image}')
                raise IboxError(e)


def _do_check(istack):
    if istack.s3_files:
        _check_s3_files(istack)

    if 'AWS::ECS::TaskDefinition' in istack.t_resources:
        _check_ecr_images(istack)


def show(istack):
    _process_template(istack)
    print('')
    logger.info(f'Resolved: {istack.name}')
    print(yaml.dump(istack.r_resources))


def process(istack):
    logger.debug('Processing Template')

    try:
        _process_template(istack)
        _do_check(istack)
    except IboxError:
        raise
    except Exception as e:
        pprint(e)
        logger.warning('Error resolving template. '
                       'Will not be able to validate s3 files and ecr images.')



from . import cfg, stacks, i_stack, i_region, events, table, ssm
from .tools import concurrent_exec, get_exports, show_confirm
from .common import *


def create():
    name = cfg.stack[0]
    cfg.exports = get_exports()
    stack = i_stack.ibox_stack(name, {})
    result = stack.create()
    if result:
        pprint(result)

    return result


def update():
    w_stacks = stacks.get()
    cfg.stacks = list(w_stacks.keys())
    cfg.exports = get_exports()
    if len(w_stacks) > 1 and (cfg.role or cfg.type) and not cfg.dryrun:
        print('You are going to UPDATE the following stacks:')
        print(cfg.stacks)
        if not show_confirm():
            return
    result = concurrent_exec('update', w_stacks, i_stack)
    if not cfg.dryrun:
        pprint(result)

    return result


def delete():
    w_stacks = stacks.get()
    cfg.stacks = list(w_stacks.keys())
    print('You are going to DELETE the following stacks:')
    print(cfg.stacks)
    if not show_confirm():
        return
    result = concurrent_exec('delete', w_stacks, i_stack)
    pprint(result)

    return result


def cancel_update():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec('cancel_update', w_stacks, i_stack)
    pprint(result)

    return result


def continue_update():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec('continue_update', w_stacks, i_stack)
    pprint(result)

    return result


def parameters():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec('parameters', w_stacks, i_stack)


def info():
    if not cfg.compact:
        cfg.OUT_WIDTH = 80
    w_stacks = stacks.get()
    result = concurrent_exec('info', w_stacks, i_stack)


def log():
    name = cfg.stack[0]
    stack = i_stack.ibox_stack(name, {})
    stack.log()


def resolve():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec('resolve', w_stacks, i_stack)

    return result


def dash():
    w_stacks = stacks.get()
    cfg.dash_name = '_' + '_'.join(cfg.stack)
    cfg.jobs = 1
    result = concurrent_exec('dash', w_stacks, i_stack)

    return result


def show_table():
    w_stacks = stacks.get()
    s_table = table.get(list(w_stacks.values()))
    print(s_table)

    return s_table


def ssm_setup():
    w_stacks = stacks.get(exit_if_empty=False)
    result = concurrent_exec(
        'ssm_setup', {k: w_stacks for k in cfg.regions}, i_region)
    pprint(result)


def ssm_put():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    stacks_params = concurrent_exec(
        'parameters', w_stacks, i_stack, **{'check': True})
    regions = ssm.get_setupped_regions()
    w_regions = cfg.regions if cfg.regions else regions
    result = concurrent_exec(
        'ssm_put', {k: stacks_params for k in w_regions if k in regions},
        i_region)
    pprint(result)


def ssm_show():
    w_stacks = stacks.get()
    regions = ssm.get_setupped_regions()
    result = concurrent_exec(
        'ssm_get', {k: w_stacks for k in regions}, i_region)
    result = ssm.show(result)
    print(result)


def r53():
    w_stacks = stacks.get()
    cfg.exports = get_exports()
    result = concurrent_exec('r53', w_stacks, i_stack)
    pprint(result)


def replicate():
    regions = ssm.get_setupped_regions() if not cfg.regions else cfg.regions

    if cfg.no_replicate_current:
        try:
            regions.remove(cfg.boto3.region_name)
        except Exception as e:
            pass

    result = concurrent_exec(
        'replicate', {k: {} for k in regions}, i_region)
    pprint(result)

    return result



def set_changed(istack):
    istack.after['resources'] = get(istack)

    before = istack.before['resources']
    after = istack.after['resources']

    changed = {}
    for r, v in before.items():
        if r in after and v != after[r]:
            changed[r] = after[r]

    istack.changed['resources'] = changed


def get(istack, dash=None):
    resources = {}
    res_list = list(istack.cfg.RESOURCES_MAP.keys())

    paginator = istack.client.get_paginator('list_stack_resources')
    response_iterator = paginator.paginate(StackName=istack.name)

    for r in response_iterator:
        for res in r['StackResourceSummaries']:
            res_lid = res['LogicalResourceId']
            res_type = res['ResourceType']
            if res_lid in res_list:
                res_pid = res['PhysicalResourceId']
                if res_pid.startswith('arn'):
                    res_pid = res_pid.split(':', 5)[5]
                if res_lid in [
                        'ListenerHttpsExternalRules1',
                        'ListenerHttpsExternalRules2',
                        'ListenerHttpExternalRules1',
                        'ListenerHttpInternalRules1']:
                    res_pid = '/'.join(res_pid.split('/')[1:4])
                if res_lid == 'ScalableTarget':
                    res_pid = res_pid.split('/')[1]
                if res_lid == 'Service':
                    res_pid_arr = res_pid.split('/')
                    if len(res_pid_arr) == 3:
                        res_pid = res_pid_arr[2]
                    else:
                        res_pid = res_pid_arr[1]
                if res_lid in [
                        'LoadBalancerApplicationExternal',
                        'LoadBalancerApplicationInternal']:
                    res_pid = '/'.join(res_pid.split('/')[1:4])

                if dash and istack.cfg.RESOURCES_MAP[res_lid]:
                    res_lid = istack.cfg.RESOURCES_MAP[res_lid]

                resources[res_lid] = res_pid

    return resources



from . import cfg
from .tools import stack_resource_to_dict
from .common import *


def show_changed(istack):
    before = istack.before['outputs']
    after = get(istack.stack)

    changed = {}
    for o, v in after.items():
        if o in before and v != before[o]:
            changed[o] = f'{before[o]} => {v}'
        elif o not in before:
            changed[o] = v

    istack.changed['outputs'] = changed
    istack.cfg.OUT_WIDTH = 80
    show(istack, 'changed')


# show stack current outputs as dict
def show(istack, when):
    outputs = getattr(istack, when)['outputs']

    out = pformat(outputs, width=istack.cfg.OUT_WIDTH)

    istack.mylog(f'{when.upper()} - STACK OUTPUTS\n{out}\n')


def get(stack):
    outputs = {}

    try:
        stack['StackName']
    except Exception:
        stack = stack_resource_to_dict(stack)

    try:
        s_outputs = stack['Outputs']
    except Exception:
        pass
    else:
        for output in s_outputs:
            key = output['OutputKey']
            value = output.get('OutputValue', None)
            outputs[key] = value

    for d in cfg.STACK_BASE_DATA:
        value = stack.get(d, None)
        if isinstance(value, datetime):
            value = value.strftime('%Y-%m-%d %X %Z')
        outputs[d] = value

    return outputs



from . import resources
from .common import *


def add_stack(istack):
    widget_width = {
        'stack': 12,
        'global': 24,
    }

    widget_height = {
        'stack': 6,
        'global': 6,
    }

    widget_title = {
        '5xx': '5xx - 4xx',
        'req': 'Requests - Healthy',
        'net': 'NetworkIN - NetworkOUT',
        '5xx_elb': '5xx - 4xx [ELB]',
        '50x_elb': '50x External - 50x Internal [ELB]',
    }

    widget_map = {
        'req': ['requests', 'healthy'],
        '5xx': ['5xx', '4xx'],
        '5xx_elb': ['5xx_elb', '4xx_elb'],
        '50x_elb': ['500_elb', '502_elb', '503_elb', '504_elb'],
        'net': ['netin', 'netout'],
    }

    ScalingPolicyTrackingsCpuBaseLabel = 'ScalingPolicyTrackingsCpu'
    widget_label = {}

    def get_alarm(res):
        alarms = {}
        cloudwatch = istack.boto3.resource('cloudwatch')
        for a in ['AlarmCPUHigh', 'AlarmCPULow']:
            alarm = cloudwatch.Alarm(res[a])
            alarms[a] = alarm.threshold

        return alarms['AlarmCPUHigh'], alarms['AlarmCPULow']

    def get_policy_ec2(res):
        polname = res['ScalingPolicyTrackings1'].split('/')[2]
        AutoScalingGroupName = ('AutoScalingGroupSpotName' if
                                'AutoScalingGroupSpotName' in res else
                                'AutoScalingGroupName')
        client = istack.boto3.client('autoscaling')
        response = client.describe_policies(
            AutoScalingGroupName=res[AutoScalingGroupName],
            PolicyNames=[polname],
        )

        return response['ScalingPolicies'][0]['TargetTrackingConfiguration']

    def get_policy_ecs(res):
        resname = '/'.join(
            res['ScalingPolicyTrackings1'].split('/')[2:5]).split(':')[0]
        client = istack.boto3.client('application-autoscaling')
        response = client.describe_scaling_policies(
            PolicyNames=list(istack.cfg.SCALING_POLICY_TRACKINGS_NAMES.keys()),
            ResourceId=resname,
            ServiceNamespace='ecs',
        )

        return response['ScalingPolicies'][0][
            'TargetTrackingScalingPolicyConfiguration']

    def get_policy(res):
        if 'AutoScalingGroupName' in res:
            conf = get_policy_ec2(res)
        else:
            conf = get_policy_ecs(res)

        stat = 'Average'
        value = conf['TargetValue']
        if ('CustomizedMetricSpecification' in conf and
                'Statistic' in conf['CustomizedMetricSpecification']):
            stat = conf['CustomizedMetricSpecification']['Statistic']

        return value, f'{ScalingPolicyTrackingsCpuBaseLabel}{stat}'

    def resolve_widget_map(name):
        count = 0
        for n in widget_map[name]:
            count += len(metrics[n])

        if count > 0:
            return True

    def get_widget_annotations(res):
        AlarmCPUHighThreshold = 60
        AlarmCPULowThreshold = 30
        ScalingPolicyTrackingsCpuValue = 80
        ScalingPolicyTrackingsCpuLabel = ScalingPolicyTrackingsCpuBaseLabel
        widget_annotations_type = None

        if 'AlarmCPUHigh' and 'AlarmCPULow' in res:
            AlarmCPUHighThreshold, AlarmCPULowThreshold = get_alarm(res)
            widget_annotations_type = 'step'

        if any(k in res for k in istack.cfg.SCALING_POLICY_TRACKINGS_NAMES):
            ScalingPolicyTrackingsCpuValue, ScalingPolicyTrackingsCpuLabel = (
                get_policy(res))
            widget_annotations_type = 'tracking'

        widget_annotations = {
            'tracking': [
                {
                    'value': 100
                },
                {
                    'label': ScalingPolicyTrackingsCpuLabel,
                    'value': ScalingPolicyTrackingsCpuValue,
                    'color': '#1f77b4',
                }
            ],
            'step': [
                {
                    'label': 'AlarmCPUHighThreshold',
                    'value': AlarmCPUHighThreshold,
                },
                {
                    'label': 'AlarmCPULowThreshold',
                    'value': AlarmCPULowThreshold,
                    'color': '#2ca02c',
                }
            ]
        }

        return widget_annotations, widget_annotations_type

    def do_label_exist(w_label, w_metrics):
        for index, metric in enumerate(w_metrics):
            if isinstance(metric, dict) and w_label in list(metric.values()):
                return True
            for m in metric:
                if isinstance(m, dict) and w_label in list(m.values()):
                    return index

        return None

    def do_insert_metrics(label, metric, widget, msg):
        widget_metrics = widget['properties']['metrics']
        if do_label_exist(label, metric):
            label_index = do_label_exist(label, widget_metrics)
            if label_index is None:
                widget_metrics.append(metric)
                out_msg = 'Added'
            else:
                widget_metrics[label_index] = metric
                out_msg = '-- Updated'
            if not istack.cfg.silent:
                print(f'\tMetrics: {msg} {out_msg}')

    def get_widget_map_position(wtype, index):
        if wtype == 'stack':
            # even index (left side)
            if (index & 1) == 0:
                return (0, widget_height['stack'] * index)
            # odd index (right side)
            else:
                return (widget_width['stack'], widget_height['stack'] * index)
        # global type
        else:
            return (0, widget_height['global'] * index)

    def add_annotations(w, atype):
        if 'annotations' not in w['properties']:
            w['properties']['annotations'] = {}

        annotations = w['properties']['annotations']
        # Vertical
        if atype == 'vertical':
            value_now = datetime.utcnow().strftime('%Y-%m-%dT%X.000Z')
            w_ann_vertical = {
                'label': '',
                'value': value_now,
            }
            if istack.cfg.vertical in ['after', 'before']:
                w_ann_vertical['fill'] = istack.cfg.vertical

            # vertical annotation not present
            if 'vertical' not in annotations:
                annotations['vertical'] = []
            # append only if empty or there is not annotations
            # with value in the same minute
            if (not annotations['vertical'] or
                    all(value_now[0:16] not in a['value']
                        for a in annotations['vertical'])):
                annotations['vertical'].append(w_ann_vertical)

        # Horizontal annotations
        annotations['horizontal'] = widget_annotations.get(
            widget_annotations_type, [])

        # Horizontal (ec2-ecs)
        # if atype == 'ecs':
        #    annotations['horizontal'] = widget_annotations['ecs']
        # if atype == 'ec2':
        #    annotations['horizontal'] = widget_annotations['ec2']

    def get_widget_base(wtype, wlist, windex, title, w):
        widget = {}
        widget.update({
            'type': 'metric',
            'x': get_widget_map_position(wtype, windex)[0],
            'y': get_widget_map_position(wtype, windex)[1],
            'width': widget_width[wtype],
            'height': widget_height[wtype],
            'properties': {
                'view': 'timeSeries',
                'stacked': False,
                'metrics': [],
                'region': istack.boto3.region_name,
                'title': title,
                'period': 300,
            }
        })

        # If widget already exist get metrics from current one
        if len(wlist) > 0:
            widget['properties']['metrics'] = (w[windex]['properties']
                                               ['metrics'])
            # and if exists, annotations too..
            if 'annotations' in w[windex]['properties']:
                widget['properties']['annotations'] = w[windex][
                    'properties']['annotations']
            del w[windex]
            out_msg = 'Updated'
        else:
            out_msg = 'Added'

        if istack.cfg.vertical:
            add_annotations(widget, 'vertical')

        w.insert(windex, widget)
        if not istack.cfg.silent:
            print(f'Widget:{title} {out_msg}')

        return widget

    def update_widget_stack_properties(widget, res):
        for m in metrics['cpu']:
            do_insert_metrics(widget_label['cpu'], m, widget, 'Cpu')
            do_insert_metrics(widget_label['cpu'] + 'Spot', m, widget, 'Cpu')
            # Always add cpu maximum
            do_insert_metrics('Cpu - Maximum', m, widget, 'Cpu - Maximum')
        for m in metrics['response']:
            do_insert_metrics(widget_label['response'], m, widget, 'Response')
            do_insert_metrics(
                widget_label['response_external'], m, widget,
                'Response External')
            do_insert_metrics(
                widget_label['response_internal'], m, widget,
                'Response Internal')

        add_annotations(widget, istack.StackType)

    def update_widget_5xx_properties(widget, res):
        for m in metrics['5xx']:
            do_insert_metrics(widget_label['5xx'], m, widget, '5xx')
            do_insert_metrics(
                widget_label['5xx_external'], m, widget, '5xx External')
            do_insert_metrics(
                widget_label['5xx_internal'], m, widget, '5xx Internal')

        for m in metrics['4xx']:
            do_insert_metrics(widget_label['4xx'], m, widget, '4xx')
            do_insert_metrics(
                widget_label['4xx_external'], m, widget, '4xx External')
            do_insert_metrics(
                widget_label['4xx_internal'], m, widget, '4xx Internal')

    def update_widget_5xx_elb_properties(widget, res):
        for m in metrics['5xx_elb']:
            do_insert_metrics(
                widget_label['5xx_external'], m, widget, '5xx External ELB')
            do_insert_metrics(
                widget_label['5xx_internal'], m, widget, '5xx Internal ELB')

        for m in metrics['4xx_elb']:
            do_insert_metrics(
                widget_label['4xx_external'], m, widget, '4xx External ELB')
            do_insert_metrics(
                widget_label['4xx_internal'], m, widget, '4xx Internal ELB')

    def update_widget_50x_elb_properties(widget, res):
        for m in metrics['500_elb']:
            do_insert_metrics(
                widget_label['500_external'], m, widget, '500 External ELB')
            do_insert_metrics(
                widget_label['500_internal'], m, widget, '500 Internal ELB')
        for m in metrics['502_elb']:
            do_insert_metrics(
                widget_label['502_external'], m, widget, '502 External ELB')
            do_insert_metrics(
                widget_label['502_internal'], m, widget, '502 Internal ELB')
        for m in metrics['503_elb']:
            do_insert_metrics(
                widget_label['503_external'], m, widget, '503 External ELB')
            do_insert_metrics(
                widget_label['503_internal'], m, widget, '503 Internal ELB')
        for m in metrics['504_elb']:
            do_insert_metrics(
                widget_label['504_external'], m, widget, '504 External ELB')
            do_insert_metrics(
                widget_label['504_internal'], m, widget, '504 Internal ELB')

    def update_widget_req_properties(widget, res):
        for m in metrics['requests']:
            do_insert_metrics(widget_label['req'], m, widget, 'Requests')
            do_insert_metrics(
                widget_label['req_external'], m, widget, 'Requests External')
            do_insert_metrics(
                widget_label['req_internal'], m, widget, 'Requests Internal')

        for m in metrics['healthy']:
            do_insert_metrics(widget_label['healthy'], m, widget, 'Healthy')

    def update_widget_network_properties(widget, res):
        for m in metrics['netin']:
            do_insert_metrics(widget_label['netin'], m, widget, 'NetworkIN')

        for m in metrics['netout']:
            do_insert_metrics(widget_label['netout'], m, widget, 'NetworkOUT')

    def update_dashboard(res, dashboard_name):
        cw = istack.boto3.client('cloudwatch')

        if not istack.cfg.silent:
            print(dashboard_name)

        try:
            dashboard_body = cw.get_dashboard(
                DashboardName=dashboard_name)['DashboardBody']
            dash = json.loads(dashboard_body)
        except Exception as e:
            print('DashBoard do not exist, creating one..\n')
            dash = {'widgets': []}

        w = dash['widgets']

        # BEGIN cpu
        # Find the current number of widget stacks,
        # so that the next one is added at the end
        len_stacks = len([n for n in w if n['width'] == widget_width['stack']])

        list_role = [n for n, v in enumerate(w)
                     if v['properties']['title'] == widget_title['role']]
        index_role = list_role[0] if len(list_role) > 0 else len_stacks

        widget = get_widget_base(
            'stack', list_role, index_role, widget_title['role'], w)
        update_widget_stack_properties(widget, res)

        # If metrics are empty delete widget (Ex ecs-alb)
        if len(widget['properties']['metrics']) == 0:
            del w[index_role]

        # BEGIN requests
        list_req = [n for n, v in enumerate(w)
                    if v['properties']['title'] == widget_title['req']]
        index_req = list_req[0] if len(list_req) > 0 else len(w)

        if resolve_widget_map('req'):
            widget = get_widget_base(
                'global', list_req, index_req, widget_title['req'], w)
            update_widget_req_properties(widget, res)

        # BEGIN 5xx
        list_5xx = [n for n, v in enumerate(w)
                    if v['properties']['title'] == widget_title['5xx']]
        index_5xx = list_5xx[0] if len(list_5xx) > 0 else len(w)

        if resolve_widget_map('5xx'):
            widget = get_widget_base(
                'global', list_5xx, index_5xx, widget_title['5xx'], w)
            update_widget_5xx_properties(widget, res)

        # BEGIN 5xx ELB
        list_5xx_elb = [n for n, v in enumerate(w)
                        if v['properties']['title'] == widget_title['5xx_elb']]
        index_5xx_elb = list_5xx_elb[0] if len(list_5xx_elb) > 0 else len(w)

        if resolve_widget_map('5xx_elb') and 'ServiceName' not in res:
            widget = get_widget_base(
                'global', list_5xx_elb, index_5xx_elb,
                widget_title['5xx_elb'], w)
            update_widget_5xx_elb_properties(widget, res)

        # BEGIN 50x ELB
        list_50x_elb = [n for n, v in enumerate(w)
                        if v['properties']['title'] == widget_title['50x_elb']]
        index_50x_elb = list_50x_elb[0] if len(list_50x_elb) > 0 else len(w)

        if resolve_widget_map('50x_elb') and 'ServiceName' not in res:
            widget = get_widget_base(
                'global', list_50x_elb, index_50x_elb,
                widget_title['50x_elb'], w)
            update_widget_50x_elb_properties(widget, res)

        # BEGIN network
        list_net = [n for n, v in enumerate(w)
                    if v['properties']['title'] == widget_title['net']]
        index_net = list_net[0] if len(list_net) > 0 else len(w)

        if resolve_widget_map('net'):
            widget = get_widget_base(
                'global', list_net, index_net, widget_title['net'], w)
            update_widget_network_properties(widget, res)

        # END Widgets

        if istack.cfg.debug:
            print(json.dumps(dash, indent=4))
            return

        # Put DashBoard
        out = cw.put_dashboard(
            DashboardName=dashboard_name,
            DashboardBody=json.dumps(dash, separators=(',', ':'))
        )

        if len(out['DashboardValidationMessages']) > 0:
            pprint(out)
        else:
            print('')
            mylog('CloudWatch-DashBoard[' + dashboard_name + '] Updated:')

        return True

    def mylog(string):
        print(istack.name + ' # ' + string)

    def get_metrics(res):
        # update widget_title and widget_label
        widget_title['role'] = f'{istack.EnvRole}.{istack.name}'
        title_role = widget_title['role']

        widget_label['cpu'] = f'Cpu - {istack.cfg.statistic}'
        widget_label['response'] = f'Response - {istack.cfg.statisticresponse}'
        widget_label['response_external'] = (
            f'Response External - {istack.cfg.statisticresponse}')
        widget_label['response_internal'] = (
            f'Response Internal - {istack.cfg.statisticresponse}')
        widget_label['5xx'] = f'{title_role} 5xx'
        widget_label['4xx'] = f'{title_role} 4xx'
        widget_label['500_elb'] = f'{title_role} 500'
        widget_label['502_elb'] = f'{title_role} 502'
        widget_label['503_elb'] = f'{title_role} 503'
        widget_label['504_elb'] = f'{title_role} 504'
        widget_label['5xx_external'] = f'{title_role} External - 5xx'
        widget_label['4xx_external'] = f'{title_role} External - 4xx'
        widget_label['5xx_internal'] = f'{title_role} Internal - 5xx'
        widget_label['4xx_internal'] = f'{title_role} Internal - 4xx'
        widget_label['500_external'] = f'{title_role} External - 500'
        widget_label['500_internal'] = f'{title_role} Internal - 500'
        widget_label['502_external'] = f'{title_role} External - 502'
        widget_label['502_internal'] = f'{title_role} Internal - 502'
        widget_label['503_external'] = f'{title_role} External - 503'
        widget_label['503_internal'] = f'{title_role} Internal - 503'
        widget_label['504_external'] = f'{title_role} External - 504'
        widget_label['504_internal'] = f'{title_role} Internal - 504'
        widget_label['req'] = f'{title_role} - Requests'
        widget_label['healthy'] = f'{title_role} - Healthy'
        widget_label['req_external'] = f'{title_role} External - Requests'
        widget_label['req_internal'] = f'{title_role} Internal - Requests'
        widget_label['netin'] = f'{title_role} - NetworkIN'
        widget_label['netout'] = f'{title_role} - NetworkOUT'

        # Set common variable for ELB Classic and Application used by EC2 stack
        if any(n in res
               for n in ['LoadBalancerNameExternal',
                         'LoadBalancerNameInternal']):
            LoadBalancerName = 'LoadBalancerName'
            LoadBalancerNameExternal = 'LoadBalancerNameExternal'
            LoadBalancerNameInternal = 'LoadBalancerNameInternal'
            AWS_ELB = 'AWS/ELB'
            Latency = 'Latency'
            HTTPCode_Backend_5XX = 'HTTPCode_Backend_5XX'
            HTTPCode_Backend_4XX = 'HTTPCode_Backend_4XX'
            HTTPCode_ELB_5XX = 'HTTPCode_ELB_5XX'
            HTTPCode_ELB_4XX = 'HTTPCode_ELB_4XX'
        elif any(n in res
                 for n in ['LoadBalancerExternal', 'LoadBalancerInternal']):
            LoadBalancerName = 'LoadBalancer'
            LoadBalancerNameExternal = 'LoadBalancerExternal'
            LoadBalancerNameInternal = 'LoadBalancerInternal'
            AWS_ELB = 'AWS/ApplicationELB'
            Latency = 'TargetResponseTime'
            HTTPCode_Backend_5XX = 'HTTPCode_Target_5XX_Count'
            HTTPCode_Backend_4XX = 'HTTPCode_Target_4XX_Count'
            HTTPCode_ELB_5XX = 'HTTPCode_ELB_5XX_Count'
            HTTPCode_ELB_4XX = 'HTTPCode_ELB_4XX_Count'
        else:
            LoadBalancerName = None
            LoadBalancerNameExternal = None
            LoadBalancerNameInternal = None
            AWS_ELB = None
            Latency = None
            HTTPCode_Backend_5XX = None
            HTTPCode_Backend_4XX = None

        metrics = {
            'cpu': [],
            'response': [],
            '5xx': [],
            '4xx': [],
            '5xx_elb': [],
            '4xx_elb': [],
            '500_elb': [],
            '502_elb': [],
            '503_elb': [],
            '504_elb': [],
            'requests': [],
            'healthy': [],
            'netin': [],
            'netout': []
        }

        # ECS
        if all(n in res for n in ['ServiceName', 'ClusterName']):
            # CPU
            metrics['cpu'].append([
                'AWS/ECS',
                'CPUUtilization',
                'ServiceName',
                res['ServiceName'],
                'ClusterName',
                res['ClusterName'],
                {
                    'period': 300,
                    'stat': istack.cfg.statistic,
                    'label': widget_label['cpu']
                }
            ])
            # Always add cpu maximum
            metrics['cpu'].append([
                'AWS/ECS',
                'CPUUtilization',
                'ServiceName',
                res['ServiceName'],
                'ClusterName',
                res['ClusterName'],
                {
                    'period': 300,
                    'stat': 'Maximum',
                    'label': 'Cpu - Maximum'
                }
            ])
            # TargetGroupExternal
            if all(n in res
                    for n in ['TargetGroupExternal', 'LoadBalancerExternal']):
                # Response Time
                metrics['response'].append([
                    'AWS/ApplicationELB',
                    'TargetResponseTime',
                    'TargetGroup',
                    res['TargetGroupExternal'],
                    'LoadBalancer',
                    res['LoadBalancerExternal'],
                    {
                        'period': 300,
                        'stat': istack.cfg.statisticresponse,
                        'yAxis': 'right',
                        'label': widget_label['response_external']
                    }
                ])
                # Healthy
                metrics['healthy'].append([
                    'AWS/ApplicationELB',
                    'HealthyHostCount',
                    'TargetGroup',
                    res['TargetGroupExternal'],
                    'LoadBalancer',
                    res['LoadBalancerExternal'],
                    {
                        'label': widget_label['healthy'],
                        'stat': istack.cfg.statistic,
                        'yAxis': 'right'
                    }
                ])
                # Requests
                metrics['requests'].append([
                    'AWS/ApplicationELB',
                    'RequestCount',
                    'TargetGroup',
                    res['TargetGroupExternal'],
                    'LoadBalancer',
                    res['LoadBalancerExternal'],
                    {
                        'label': widget_label['req_external'],
                        'stat': 'Sum'
                    }
                ])
                # 5xx
                metrics['5xx'].append([
                    'AWS/ApplicationELB',
                    'HTTPCode_Target_5XX_Count',
                    'TargetGroup',
                    res['TargetGroupExternal'],
                    'LoadBalancer',
                    res['LoadBalancerExternal'],
                    {
                        'label': widget_label['5xx_external'],
                        'stat': 'Sum'
                    }
                ])
                # 4xx
                metrics['4xx'].append([
                    'AWS/ApplicationELB',
                    'HTTPCode_Target_4XX_Count',
                    'TargetGroup',
                    res['TargetGroupExternal'],
                    'LoadBalancer',
                    res['LoadBalancerExternal'],
                    {
                        'label': widget_label['4xx_external'],
                        'stat': 'Sum',
                        'yAxis': 'right'
                    }
                ])
            # TargetGroupInternal
            if all(n in res
                    for n in ['TargetGroupInternal', 'LoadBalancerInternal']):
                # Response Time
                metrics['response'].append([
                    'AWS/ApplicationELB',
                    'TargetResponseTime',
                    'TargetGroup',
                    res['TargetGroupInternal'],
                    'LoadBalancer',
                    res['LoadBalancerInternal'],
                    {
                        'period': 300,
                        'stat': istack.cfg.statisticresponse,
                        'yAxis': 'right',
                        'label': widget_label['response_internal']
                    }
                ])
                # Healthy
                metrics['healthy'].append([
                    'AWS/ApplicationELB',
                    'HealthyHostCount',
                    'TargetGroup',
                    res['TargetGroupInternal'],
                    'LoadBalancer',
                    res['LoadBalancerInternal'],
                    {
                        'label': widget_label['healthy'],
                        'stat': istack.cfg.statistic,
                        'yAxis': 'right'
                    }
                ])
                # Requests
                metrics['requests'].append([
                    'AWS/ApplicationELB',
                    'RequestCount',
                    'TargetGroup',
                    res['TargetGroupInternal'],
                    'LoadBalancer',
                    res['LoadBalancerInternal'],
                    {
                        'label': widget_label['req_internal'],
                        'stat': 'Sum'
                    }
                ])
                # 5xx
                metrics['5xx'].append([
                    'AWS/ApplicationELB',
                    'HTTPCode_Target_5XX_Count',
                    'TargetGroup',
                    res['TargetGroupInternal'],
                    'LoadBalancer',
                    res['LoadBalancerInternal'],
                    {
                        'label': widget_label['5xx_internal'],
                        'stat': 'Sum'
                    }
                ])
                # 4xx
                metrics['4xx'].append([
                    'AWS/ApplicationELB',
                    'HTTPCode_Target_4XX_Count',
                    'TargetGroup',
                    res['TargetGroupInternal'],
                    'LoadBalancer',
                    res['LoadBalancerInternal'],
                    {
                        'label': widget_label['4xx_internal'],
                        'stat': 'Sum',
                        'yAxis': 'right'
                    }
                ])
        else:
            # EC2
            if all(n in res for n in ['AutoScalingGroupName']):
                # CPU
                metrics['cpu'].append([
                    'AWS/EC2',
                    'CPUUtilization',
                    'AutoScalingGroupName',
                    res['AutoScalingGroupName'],
                    {
                        'period': 300,
                        'stat': istack.cfg.statistic,
                        'label': widget_label['cpu']
                    }
                ])
                # Always add cpu maximum
                metrics['cpu'].append([
                    'AWS/EC2',
                    'CPUUtilization',
                    'AutoScalingGroupName',
                    res['AutoScalingGroupName'],
                    {
                        'period': 300,
                        'stat': 'Maximum',
                        'label': 'Cpu - Maximum'
                    }
                ])
                # CPU Spot
                if all(n in res for n in ['AutoScalingGroupSpotName']):
                    metrics['cpu'].append([
                        'AWS/EC2',
                        'CPUUtilization',
                        'AutoScalingGroupName',
                        res['AutoScalingGroupSpotName'],
                        {
                            'period': 300,
                            'stat': istack.cfg.statistic,
                            'label': widget_label['cpu'] + 'Spot'
                        }
                    ])
                # Healthy
                metrics['healthy'].append([
                    'AWS/AutoScaling',
                    'GroupInServiceInstances',
                    'AutoScalingGroupName',
                    res['AutoScalingGroupName'],
                    {
                        'label': widget_label['healthy'],
                        'stat': istack.cfg.statistic,
                        'yAxis': 'right'
                    }
                ])
                # Network
                metrics['netin'].append([
                    'AWS/EC2',
                    'NetworkIn',
                    'AutoScalingGroupName',
                    res['AutoScalingGroupName'],
                    {
                        'label': widget_label['netin'],
                        'period': 300,
                        'stat': 'Sum'
                    }
                ])
                metrics['netout'].append([
                    'AWS/EC2',
                    'NetworkOut',
                    'AutoScalingGroupName',
                    res['AutoScalingGroupName'],
                    {
                        'label': widget_label['netout'],
                        'period': 300,
                        'stat': 'Sum',
                        'yAxis': 'right'
                    }
                ])

            # ELB
            for n in ['External', 'Internal']:
                res_name = locals()[f'LoadBalancerName{n}']
                label = n.lower()

                if res_name in res:
                    # Response
                    metrics['response'].append([
                        AWS_ELB,
                        Latency,
                        LoadBalancerName,
                        res[res_name],
                        {
                            'period': 300,
                            'stat': istack.cfg.statisticresponse,
                            'yAxis': 'right',
                            'label': widget_label[f'response_{label}']
                        }
                    ])
                    # Requests
                    metrics['requests'].append([
                        AWS_ELB,
                        'RequestCount',
                        LoadBalancerName,
                        res[res_name],
                        {
                            'label': widget_label[f'req_{label}'],
                            'stat': 'Sum'
                        }
                    ])
                    # 5xx
                    metrics['5xx'].append([
                        AWS_ELB,
                        HTTPCode_Backend_5XX,
                        LoadBalancerName,
                        res[res_name],
                        {
                            'label': widget_label[f'5xx_{label}'],
                            'stat': 'Sum'
                        }
                    ])
                    # 4xx
                    metrics['4xx'].append([
                        AWS_ELB,
                        HTTPCode_Backend_4XX,
                        LoadBalancerName,
                        res[res_name],
                        {
                            'label': widget_label[f'4xx_{label}'],
                            'stat': 'Sum',
                            'yAxis': 'right'
                        }
                    ])
                    # 5xx ELB
                    metrics['5xx_elb'].append([
                        AWS_ELB,
                        HTTPCode_ELB_5XX,
                        LoadBalancerName,
                        res[res_name],
                        {
                            'label': widget_label[f'5xx_{label}'],
                            'stat': 'Sum'
                        }
                    ])
                    # 4xx ELB
                    metrics['4xx_elb'].append([
                        AWS_ELB,
                        HTTPCode_ELB_4XX,
                        LoadBalancerName,
                        res[res_name],
                        {
                            'label': widget_label[f'4xx_{label}'],
                            'stat': 'Sum',
                            'yAxis': 'right'
                        }
                    ])

                    # 50x ELB
                    if f'LoadBalancer{n}' in res:
                        # 500
                        metrics['500_elb'].append([
                            'AWS/ApplicationELB',
                            'HTTPCode_ELB_500_Count',
                            'LoadBalancer',
                            res[res_name],
                            {
                                'label': widget_label[f'500_{label}'],
                                'stat': 'Sum',
                                'yAxis': 'right' if n == 'Internal' else 'left'
                            }
                        ])
                        # 502
                        metrics['502_elb'].append([
                            'AWS/ApplicationELB',
                            'HTTPCode_ELB_502_Count',
                            'LoadBalancer',
                            res[res_name],
                            {
                                'label': widget_label[f'502_{label}'],
                                'stat': 'Sum',
                                'yAxis': 'right' if n == 'Internal' else 'left'
                            }
                        ])
                        # 503
                        metrics['503_elb'].append([
                            'AWS/ApplicationELB',
                            'HTTPCode_ELB_503_Count',
                            'LoadBalancer',
                            res[res_name],
                            {
                                'label': widget_label[f'503_{label}'],
                                'stat': 'Sum',
                                'yAxis': 'right' if n == 'Internal' else 'left'
                            }
                        ])
                        # 504
                        metrics['504_elb'].append([
                            'AWS/ApplicationELB',
                            'HTTPCode_ELB_504_Count',
                            'LoadBalancer',
                            res[res_name],
                            {
                                'label': widget_label[f'504_{label}'],
                                'stat': 'Sum',
                                'yAxis': 'right' if n == 'Internal' else 'left'
                            }
                        ])

        return metrics

    # get stack resources
    res = resources.get(istack, dash=True)

    # get widget annotations and widget_annotations_type for alarms threshold
    # or policy tracking target value
    widget_annotations, widget_annotations_type = get_widget_annotations(res)

    if not istack.cfg.silent:
        pprint(res)

    # get metrics
    metrics = get_metrics(res)

    print('')
    # update dashboards
    update_dashboard(res, istack.cfg.dash_name)


def update(istack):
    cw_client = istack.boto3.client('cloudwatch')
    response_dash = cw_client.list_dashboards(DashboardNamePrefix='_')

    resources.set_changed(istack)
    if istack.cfg.dashboard == 'OnChange':
        res_changed = istack.changed['resources']
        mode = ''
    elif istack.cfg.dashboard in ['Always', 'Generic']:
        res_changed = istack.after['resources']
        mode = istack.cfg.dashboard
    else:
        return

    if (res_changed
            or any(n in istack.changed['outputs'] for n in [
                'ScalingPolicyTrackings',
                'AutoScalingScalingPolicy',
                'ApplicationAutoScalingScalingPolicy'])):
        for dash in response_dash['DashboardEntries']:
            if istack.name in dash['DashboardName']:
                istack.cfg.dash_name = dash['DashboardName']
                add_stack(istack)



from . import cfg, parameters, outputs
from .aws import myboto3
from .log import logger
from .common import *


def get_base_data(stack):
    data = {
        'before': {},
        'after': {},
        'changed': {},
        'parameter_not_empty': [],
    }

    stack_outputs = outputs.get(stack)

    data.update(stack_outputs)
    data['before']['outputs'] = stack_outputs

    stack_parameters = parameters.get(stack)
    if stack_parameters:
        data['c_parameters'] = stack_parameters

        # add parameters too, this way i can show them using show command
        for n, v in data['c_parameters'].items():
            if (not n.startswith('Env')
                    and n not in ['UpdateMode']
                    and v != '' and ',,,' not in v):
                data['parameter_not_empty'].append(n)
            if n not in data:
                data[n] = v

    return data


def _get_stack(r, data):
    for s in r['Stacks']:
        stack_name = s['StackName']
        stack_data = get_base_data(s)
        stack_role = stack_data.get('EnvRole', None)
        stack_type = stack_data.get('StackType', None)
        if (stack_name in cfg.stack
                or stack_role in cfg.role
                or stack_type in cfg.type
                or 'ALL' in cfg.type):
            data[stack_name] = stack_data


def get(names=[], exit_if_empty=True, obj=None):
    if not obj:
        boto3 = myboto3()
        client = boto3.client('cloudformation')
    else:
        boto3 = getattr(obj, 'boto3')
        client = boto3.client('cloudformation')

    logger.info('Getting Stacks Description')
    data = {}

    if not names:
        names = cfg.stack

    if (not cfg.role
            and not cfg.type
            and len(names) < cfg.MAX_SINGLE_STACKS):
        for s in names:
            response = client.describe_stacks(StackName=s)
            _get_stack(response, data)
    else:
        paginator = client.get_paginator('describe_stacks')
        response_iterator = paginator.paginate()
        for r in response_iterator:
            _get_stack(r, data)

    if not data and exit_if_empty:
        logger.warning('No Stacks found!\n')
        exit(0)

    return data



from calendar import timegm
from .common import *


# show old and new service tasks during an update
def _show_service_update(istack, event, timedelta):
    # avoid showing current service log if is requested a stack past event
    if timedelta != '0':
        return

    service_logical_resource_id = event.logical_resource_id
    service = task = cluster = deps_before = None
    deployment_task = ''
    deployments_len = pendingCount = stuck_n = 0
    client = istack.boto3.client('ecs')

    try:
        cluster = istack.stack.Resource(
            'ScalableTarget').physical_resource_id.split('/')[1]
        service = istack.stack.Resource(
            service_logical_resource_id).physical_resource_id
    except Exception:
        return

    deps = {
        'PRIMARY': {},
        'ACTIVE': {},
    }
    while task != deployment_task or deployments_len > 1 or pendingCount != 0:
        istack.stack.reload()
        task = istack.stack.Resource('TaskDefinition').physical_resource_id
        response = client.describe_services(
            cluster=cluster,
            services=[service],
        )
        deployments = response['services'][0]['deployments']
        deployments_len = len(deployments)
        for dep in deployments:
            status = dep['status']
            for p in [
                    'desiredCount', 'runningCount',
                    'pendingCount', 'taskDefinition']:
                deps[status][p] = dep[p]

        deployment_task = deps['PRIMARY']['taskDefinition']
        desiredCount = deps['PRIMARY']['desiredCount']
        pendingCount = deps['PRIMARY']['pendingCount']
        runningCount = deps['PRIMARY']['runningCount']

        if str(deps) != deps_before:
            deps_before = str(deps)
            for d in ['PRIMARY', 'ACTIVE']:
                if 'taskDefinition' in deps[d]:
                    deps[d]['taskDefinition'] = deps[d][
                        'taskDefinition'].split('/')[-1]
            istack.mylog(
                'PRIMARY: %s' %
                pformat(
                    deps['PRIMARY'],
                    width=1000000
                )
            )
            istack.mylog(
                'ACTIVE: %s\n' %
                pformat(
                    deps['ACTIVE'],
                    width=1000000
                )
            )

            # is update stuck ?
            max_retry = istack.cfg.max_retry_ecs_service_running_count
            if max_retry > 0 and stuck_n > max_retry:
                istack.last_event_timestamp = event.timestamp
                raise IboxErrorECSService(
                    'ECS Service did not stabilize '
                    f'[{stuck_n} > {max_retry}] - '
                    'cancelling update [ROLLBACK]')

            if desiredCount > 0 and runningCount == 0:
                stuck_n += 1

        time.sleep(5)


# get timestamp from last event available
def get_last_timestamp(istack):
    last_event = list(istack.stack.events.all().limit(1))[0]

    return last_event.timestamp


# show all events after specific timestamp and return last event timestamp
def show(istack, timestamp, timedelta='0'):
    event_iterator = istack.stack.events.all()
    event_list = []
    for event in event_iterator:
        if event.timestamp > timestamp:
            event_list.insert(0, event)
        else:
            break
    for event in event_list:
        logtime = timegm(event.timestamp.timetuple())
        istack.mylog(
            event.logical_resource_id +
            " " + event.resource_status +
            " " + str(datetime.fromtimestamp(logtime)) +
            " " + str(event.resource_status_reason)
        )
        if (
            event.resource_type == 'AWS::ECS::Service'
            and event.resource_status == 'UPDATE_IN_PROGRESS'
            and event.resource_status_reason is None
            and istack.stack.stack_status not in
                istack.cfg.STACK_COMPLETE_STATUS
        ):
            _show_service_update(istack, event, timedelta)

    if len(event_list) > 0:
        return(event_list.pop().timestamp)

    return timestamp



from . import resources, changeset, events, outputs, dashboard
from .tags import get_action_tags
from .tools import show_confirm
from .log import logger
from .common import *


# build all args for action
def _get_action_args(istack):
    us_args = {}
    us_args['StackName'] = istack.name
    us_args['Parameters'] = istack.action_parameters
    us_args['Tags'] = istack.action_tags
    us_args['Capabilities'] = [
        'CAPABILITY_IAM',
        'CAPABILITY_NAMED_IAM',
        'CAPABILITY_AUTO_EXPAND',
    ]

    # sns topic
    us_args['NotificationARNs'] = istack.cfg.topics

    # Handle policy during update
    if hasattr(istack.cfg, 'policy') and istack.cfg.policy:
        action = ['"Update:%s"' % a for a in istack.cfg.policy.split(',')]
        action = '[%s]' % ','.join(action)
        us_args['StackPolicyDuringUpdateBody'] = (
            '{"Statement" : [{"Effect" : "Allow",'
            '"Action" :%s,"Principal": "*","Resource" : "*"}]}' % action)

    if istack.template_from == 'Current':
        us_args['UsePreviousTemplate'] = True
    if istack.template_from == 'S3':
        us_args['TemplateURL'] = istack.cfg.template
    if istack.template_from == 'File':
        us_args['TemplateBody'] = json.dumps(istack.template)

    return us_args


# wait update until complete showing events status
def _update_waiter(istack, timestamp=None):
    last_timestamp = timestamp if timestamp else istack.last_event_timestamp

    # return without waiting
    if istack.cfg.nowait:
        istack.stack.reload()
        return

    while True:
        try:
            istack.stack.reload()
        except botocore.exceptions.ClientError as e:
            print(e)

        if istack.stack.stack_status in istack.cfg.STACK_COMPLETE_STATUS:
            events.show(istack, last_timestamp)
            break

        time.sleep(istack.cfg.ACTION_WAITER_SLEEP_TIME)

        try:
            last_timestamp = events.show(istack, last_timestamp)
        except IboxErrorECSService as e:
            # ECS Service did not stabilize, cancel update [ROLLBACK]
            logger.warning(e.args[0])
            cancel_update(istack)

    print('\n')


def create(istack):
    stack_tags = [
        {'Key': 'Env', 'Value': istack.cfg.Env},
        {'Key': 'EnvRole', 'Value': istack.cfg.EnvRole},
        {'Key': 'EnvStackVersion', 'Value': istack.cfg.version},
        {'Key': 'EnvApp1Version', 'Value': istack.cfg.EnvApp1Version},
    ]

    # set tags
    istack.action_tags = get_action_tags(istack, stack_tags)

    if not show_confirm():
        return

    # get final args for update
    us_args = _get_action_args(istack)

    response = istack.client.create_stack(**us_args)
    istack.mylog(f'{json.dumps(response)}\n')
    time.sleep(1)

    istack.stack = istack.cloudformation.Stack(istack.name)
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    _update_waiter(istack)

    return True


def update(istack):
    # set tags
    istack.action_tags = get_action_tags(istack, istack.stack.tags)

    # get final args for update
    us_args = _get_action_args(istack)

    outputs.show(istack, 'before')

    # -if using changeset ...
    if not istack.cfg.nochangeset and (
            len(istack.cfg.stacks) == 1 or istack.cfg.dryrun):
        changeset_ok = changeset.process(istack, us_args)
        if not changeset_ok:
            return

    istack.before['resources'] = resources.get(istack)
    istack.last_event_timestamp = events.get_last_timestamp(istack)

    # do stack update
    response = istack.client.update_stack(**us_args)
    istack.mylog(f'{json.dumps(response)}\n')
    time.sleep(1)

    # -show update status until complete
    _update_waiter(istack)

    # show changed outputs
    outputs.show_changed(istack)

    # update dashboard
    dashboard.update(istack)

    return True


def delete(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.stack.delete()
    istack.mylog(f'{json.dumps(response)}\n')
    # -show update status until complete
    _update_waiter(istack)

    return True


def cancel_update(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.stack.cancel_update()
    istack.mylog(f'{json.dumps(response)}\n')
    # -show update status until complete
    _update_waiter(istack)

    return True


def continue_update(istack):
    istack.last_event_timestamp = events.get_last_timestamp(istack)
    response = istack.client.continue_update_rollback(
        StackName=istack.name,
        ResourcesToSkip=istack.cfg.resources_to_skip)
    istack.mylog(f'{json.dumps(response)}\n')
    # -show update status until complete
    _update_waiter(istack)

    return True


def log(istack):
    last_timestamp = events.get_last_timestamp(istack)
    time_delta = int(istack.cfg.timedelta)

    if time_delta == 0:
        time_event = last_timestamp - timedelta(seconds=1)
        _update_waiter(istack, time_event)
    else:
        if time_delta < 30:
            time_delta = time_delta * 86400
        time_event = last_timestamp - timedelta(seconds=time_delta)
        events.show(istack, time_event, time_delta)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iboxstacksops",
    version="0.3.5",
    author="Mello",
    author_email="mello+python@ankot.org",
    description="AWS Infrastructure in a Box - Stacks management program",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mello7tre/AwsIBoxStackOps",
    packages=[
        'iboxstacksops',
    ],
    package_data={},
    install_requires=[
        'boto3',
        'prettytable',
        'PyYAML>=5,==5.*',
    ],
    extras_require = {
        'extra': ['slackclient'],
    },
    python_requires='>=3.7',
    scripts=[
        'scripts/ibox_stacksops.py',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



#!/usr/bin/env python3
from prettytable import (PrettyTable,
                         ALL as ptALL, FRAME as ptFRAME,
                         HEADER as ptHEADER, NONE as ptNONE)
from . import cfg
from .log import logger
from .common import *


def get(data):
    fields = cfg.fields
    table = PrettyTable()
    table.padding_width = 1
    table.field_names = fields
    for n in data:
        table.add_row([
            'Null' if (i not in n or not n[i]) else f'{n[i]} *'
            if i in n['parameter_not_empty'] else n[i] for i in fields])

    table.sortby = fields[0]
    table.reversesort = True
    table.align = 'l'

    if cfg.output == 'html':
        table.format = True

        return table.get_html_string(fields=fields)

    if cfg.output == 'bare':
        table.header = False
        table.border = False
        table.left_padding_width = 0
        table.padding_width = 1

    return table.get_string(fields=fields)



from . import resources
from .common import *


def create(istack):

    def _get_rec_info(record, rtype):
        r = {}
        param = record.split('.')
        r['stack'] = param[0]
        r['role'] = param[1]
        r['type'] = rtype
        if rtype == 'external':
            r['region'] = param[2]
            r['domain'] = '.'.join(param[3:6])
        if rtype == 'internal':
            r['domain'] = '.'.join(param[2:5])
        if rtype == 'cf':
            del r['stack']
            r['role'] = param[0]
            r['domain'] = '.'.join(param[2:5])

        if istack.cfg.suffix and rtype != 'cf':
            r['role'] = r['role'] + '-' + istack.cfg.suffix

        return r

    def _get_record_type(zoneid, name):
        resp = istack.route53.list_resource_record_sets(
            HostedZoneId=zoneid,
            StartRecordName=name,
            MaxItems='1',
        )

        if resp['ResourceRecordSets']:
            return resp['ResourceRecordSets'][0]['Type']
        else:
            return 'A'

    def _get_record_change(name, zoneid, target, rtype):
        changes = {
            'Action': 'UPSERT',
            'ResourceRecordSet': {
                'Name': name,
                'Type': rtype,
                'AliasTarget': {
                    'HostedZoneId': zoneid,
                    'DNSName': target,
                    'EvaluateTargetHealth': False
                }
            }
        }

        return changes

    def _get_zoneid(record):
        zones = istack.route53.list_hosted_zones_by_name(
            DNSName=record['domain'])['HostedZones']
        for z in zones:
            zoneid = z['Id'].split('/')[2]
            zone = istack.route53.get_hosted_zone(Id=zoneid)
            if zone['HostedZone']['Name'] != record['domain'] + '.':
                continue
            try:
                zone_region = zone['VPCs'][0]['VPCRegion']
            except Exception:
                return zoneid
            else:
                if zone_region == istack.boto3.region_name:
                    return zoneid

    istack.cfg.RESOURCES_MAP = istack.cfg.RESOURCES_MAP_R53
    res = resources.get(istack)
    pprint(res)
    out = {}
    for r, v in res.items():
        r_out = {}
        if r.startswith('RecordSetExternal'):
            record = _get_rec_info(v, 'external')
            record_region = '%s.%s.%s' % (
                record['role'], record['region'], record['domain'])
            record_origin = '%s.origin.%s' % (
                record['role'], record['domain'])
            record_cf = '%s.%s' % (
                record['role'], record['domain'])
            map_record = {
                record_region: v,
            }

            if 'RecordSetCloudFront' not in res:
                map_record[record_cf] = record_region

            if not istack.cfg.noorigin:
                map_record[record_origin] = record_region

        if r.startswith('RecordSetInternal'):
            record = _get_rec_info(v, 'internal')
            record_internal = record['role'] + '.' + record['domain']
            map_record = {
                record_internal: v,
            }

        if r == 'RecordSetCloudFront':
            record = _get_rec_info(v, 'cf')
            record_cf = record['role'] + '.' + record['domain']
            map_record = {
                record_cf: v,
            }

        zoneid = _get_zoneid(record)

        for name, target in map_record.items():
            rtype = _get_record_type(zoneid, target)
            changes = _get_record_change(name, zoneid, target, rtype)
            print(name)
            pprint(changes)
            print('')

            if istack.cfg.dryrun:
                continue

            resp = istack.route53.change_resource_record_sets(
                HostedZoneId=zoneid,
                ChangeBatch={'Changes': [changes]}
            )
            pprint(resp['ChangeInfo']['Status'])

            r_out[name] = target

        out[r] = r_out

    return out
